# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: URN related objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

>>> from MyCapytain.common.reference import URN, Reference, Citation

"""
from __future__ import unicode_literals

from collections import defaultdict
from past.builtins import basestring
from six import text_type as str
from builtins import \
    range, object
from copy import copy
import re


REFSDECL_SPLITTER = re.compile("/+[\*()|\sa-zA-Z0-9:\[\]@=\\\{\$'\"\.\s]+")
REFSDECL_REPLACER = re.compile("\$[0-9]+")
SUBREFERENCE = re.compile("(\w*)\[{0,1}([0-9]*)\]{0,1}", re.UNICODE)
REFERENCE_REPLACER = re.compile("(@[a-zA-Z0-9:]+){1}(=){1}([\\\$'\"?0-9]{3,6})")


class Reference(object):
    """ A reference object giving informations

    :param reference: Passage Reference part of a Urn
    :type reference: basestring
    :ivar parent: Parent Reference
    :type parent: Reference
    :ivar highest: List representation of the range member which is the highest in the hierarchy (If equal, start is returned)
    :type highest: Reference
    :ivar start: First part of the range
    :type start: Reference
    :ivar end: Second part of the range
    :type end: Reference
    :ivar list: List representation of the range. Not available for range
    :type list: list
    :ivar subreference: Word and Word counter ("Achiles", 1) representing the subreference. Not available for range
    :type subreference: (str, int)

    :Example:
        >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
        >>>    b = Reference(reference="1.1")
        >>>    Reference("1.1-2.2.2").highest == ["1", "1"]

    Reference object supports the following magic methods : len(), str() and eq().

    :Example:
        >>>    len(a) == 2 && len(b) == 1
        >>>    str(a) == "1.1@Achiles[1]-1.2@Zeus[1]"
        >>>    b == Reference("1.1") && b != a

    .. note::
        While Reference(...).subreference and .list are not available for range, Reference(..).start.subreference and Reference(..).end.subreference as well as .list are available
    """

    def __init__(self, reference=""):
        self.reference = reference
        if reference == "":
            self.parsed = (self.__model(), self.__model())
        else:
            self.parsed = self.__parse(reference)

    @property
    def parent(self):
        """ Parent of the actual URN, for example, 1.1 for 1.1.1

        :rtype: Reference
        """
        if len(self.parsed[0][1]) == 1 and len(self.parsed[1][1]) <= 1:
            return None
        else:
            if len(self.parsed[0][1]) > 1 and len(self.parsed[1][1]) == 0:
                return Reference("{0}{1}".format(
                    ".".join(list(self.parsed[0][1])[0:-1]),
                    self.parsed[0][3] or ""
                ))
            elif len(self.parsed[0][1]) > 1 and len(self.parsed[1][1]) > 1:
                first = list(self.parsed[0][1])[0:-1]
                last = list(self.parsed[1][1])[0:-1]
                if first == last and self.parsed[1][3] is None \
                    and self.parsed[0][3] is None:
                    return Reference(".".join(first))
                else:
                    return Reference("{0}{1}-{2}{3}".format(
                        ".".join(first),
                        self.parsed[0][3] or "",
                        ".".join(list(self.parsed[1][1])[0:-1]),
                        self.parsed[1][3] or ""
                    ))

    @property
    def highest(self):
        """ Return highest reference level

        For references such as 1.1-1.2.8, with different level, it can be useful to access to the highest node in the
        hierarchy. In this case, the highest level would be 1.1. The function would return ["1", "1"]

        .. note:: By default, this property returns the start level

        :rtype: Reference
        """
        if not self.end:
            return self
        elif len(self.start) < len(self.end) and len(self.start):
            return self.start
        elif len(self.start) > len(self.end) and len(self.end):
            return self.end
        elif len(self.start):
            return self.start
        return self

    @property
    def start(self):
        """ Quick access property for start list
        """
        if self.parsed[0][0] and len(self.parsed[0][0]):
            return Reference(self.parsed[0][0])

    @property
    def end(self):
        """ Quick access property for reference end list
        """
        if self.parsed[1][0] and len(self.parsed[1][0]):
            return Reference(self.parsed[1][0])

    @property
    def list(self):
        """ Return a list version of the object if it is a single passage

        .. note:: Access to start list and end list should be done through obj.start.list and obj.end.list

        :rtype: [str]
        """
        if not self.end:
            return self.parsed[0][1]

    @property
    def subreference(self):
        """ Return the subreference of a single node reference

        .. note:: Access to start and end subreference should be done through obj.start.subreference
        and obj.end.subreference

        :rtype: (str, int)
        """
        if not self.end:
            return Reference.convert_subreference(*self.parsed[0][2])

    def __len__(self):
        """ Return depth of highest reference level

        For references such as 1.1-1.2.8, or simple references such as 1.a, with different level, it can be useful to
        know the depth of the reference to access the right XPath for example. This property returns the depth of the
        highest node

        :example:
            - len(1.1) == 2
            - len(1.2.8-1.3) == 2
            - len(1-1.2) == 1

        :rtype: int
        """
        return len(self.highest.list)

    def __str__(self):
        """ Return full reference in string format

        :rtype: basestring
        :returns: String representation of Reference Object

        :Example:
            >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    b = Reference(reference="1.1")
            >>>    str(a) == "1.1@Achiles[1]-1.2@Zeus[1]"
            >>>    str(b) == "1.1"
        """
        return self.reference

    def __eq__(self, other):
        """ Equality checker for Reference object

        :param other: An object to be checked against
        :rtype: boolean
        :returns: Equality between other and self

        :Example:
            >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    b = Reference(reference="1.1")
            >>>    c = Reference(reference="1.1")
            >>>    (a == b) == False
            >>>    (c == b) == True
        """
        return (isinstance(other, self.__class__)
                and self.reference == str(other))

    def __model(self):
        """ 3-Tuple model for references
        
        First element is full text reference,
        Second is list of passage identifiers
        Third is subreference

        :returns: An empty list to model data
        :rtype: list
        """
        return [None, [], None, None]

    def __regexp(self, subreference):
        """ Split components of subreference 
        
        :param subreference: A subreference
        :type subreference: basestring
        :rtype: List.<Tuple>
        :returns: List where first element is a tuple representing different components
        """
        return SUBREFERENCE.findall(subreference)[0]

    def __parse(self, reference):
        """ Parse references informations
        
        """

        ref = reference.split("-")
        element = [self.__model(), self.__model()]
        for i in range(0, len(ref)):
            r = ref[i]
            element[i][0] = r
            subreference = r.split("@")
            if len(subreference) == 2:
                element[i][2] = self.__regexp(subreference[1])
                element[i][3] = "@" + subreference[1]
                r = subreference[0]
            element[i][1] = r.split(".")
            element[i] = tuple(element[i])
        return tuple(element)

    @staticmethod
    def convert_subreference(word, counter):
        if len(counter) and word:
            return str(word), int(counter)
        elif len(counter) == 0 and word:
            return str(word), 0
        else:
            return "", 0


class URN(object):

    """ A URN object giving all useful sections 

    :param urn: A CTS URN
    :type urn: str
    :ivar urn_namespace: Namespace of the URN
    :type urn_namespace: str
    :ivar namespace: CTS Namespace
    :type namespace: str
    :ivar textgroup: CTS Textgroup
    :type textgroup: str
    :ivar work: CTS Work
    :type work: str
    :ivar version: CTS Version
    :type version: str
    :ivar reference: CTS Reference
    :type reference: Reference
    :cvar NAMESPACE: Constant representing the URN until its namespace
    :cvar TEXTGROUP: Constant representing the URN until its textgroup
    :cvar WORK: Constant representing the URN until its work
    :cvar VERSION: Constant representing the URN until its version
    :cvar PASSAGE: Constant representing the URN until its full passage
    :cvar PASSAGE_START: Constant representing the URN until its passage (end excluded)
    :cvar PASSAGE_END: Constant representing the URN until its passage (start excluded)
    :cvar NO_PASSAGE: Constant representing the URN until its passage excluding its passage
    :cvar COMPLETE: Constant representing the complete URN

    :Example:
        >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")

    URN object supports the following magic methods : len(), str() and eq(), gt() and lt().

    :Example:
        >>>     b = URN("urn:cts:latinLit:phi1294.phi002")
        >>>     a != b
        >>>     a > b # It has more member. Only member count is compared
        >>>     b < a
        >>>     len(a) == 5 # Reference is not counted to not induce count equivalencies with the optional version
        >>>     len(b) == 4

    .. exclude-members:: all
    .. automethod:: upTo
    """

    NAMESPACE = 0
    TEXTGROUP = 1
    WORK = 2
    VERSION = 3
    PASSAGE = 4
    PASSAGE_START = 5
    PASSAGE_END = 6
    NO_PASSAGE = 10
    COMPLETE = 100

    def __init__(self, urn):
        self.__urn = None
        self.__parsed = self.__parse(urn)

    @property
    def urn_namespace(self):
        return self.__parsed["urn_namespace"]

    @urn_namespace.setter
    def urn_namespace(self, value):
        self.__urn = None
        self.__parsed["urn_namespace"] = value

    @property
    def namespace(self):
        return self.__parsed["cts_namespace"]

    @namespace.setter
    def namespace(self, value):
        self.__urn = None
        self.__parsed["cts_namespace"] = value

    @property
    def textgroup(self):
        return self.__parsed["textgroup"]

    @textgroup.setter
    def textgroup(self, value):
        self.__urn = None
        self.__parsed["textgroup"] = value

    @property
    def work(self):
        return self.__parsed["work"]

    @work.setter
    def work(self, value):
        self.__urn = None
        self.__parsed["work"] = value

    @property
    def version(self):
        return self.__parsed["version"]

    @version.setter
    def version(self, value):
        self.__urn = None
        self.__parsed["version"] = value

    @property
    def reference(self):
        return self.__parsed["reference"]

    @reference.setter
    def reference(self, value):
        self.__urn = None
        if isinstance(value, Reference):
            self.__parsed["reference"] = value
        else:
            self.__parsed["reference"] = Reference(value)

    def __len__(self):
        """ Returns the len of the URN

        :rtype: int
        :returns: Length of the URN

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    print(len(a))
        """
        
        items = [
            key
            for key, value in self.__parsed.items()
            if key not in ["reference"] and value is not None
        ]
        return len(items)

    def __gt__(self, other):
        """ Allows for greater comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of bigger size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (a > b) == True # 
        """
        return len(self) > len(other)

    def __lt__(self, other):
        """ Allows for lower comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of lower size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (b < a) == True # 
        """
        return len(self) < len(other)

    def __eq__(self, other):
        """ Equality checker for URN object
        
        :param other: An object to be checked against
        :type other: URN
        :rtype: boolean
        :returns: Equality between other and self

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (b == a) == False # 
        """
        return (isinstance(other, self.__class__)
                and self.__str__() == str(other))

    def __str__(self):
        """ Return full initial urn
        
        :rtype: basestring
        :returns: String representation of URN Object

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    str(a) == "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1"
        """
        if self.__urn is None:
            urn = "urn:" + self.__parsed["urn_namespace"]
            if self.namespace:
                urn += ":" + self.namespace
                if self.textgroup:
                    urn += ":" + self.textgroup
                    if self.work:
                        urn += "." + self.work
                        if self.version:
                            urn += "." + self.version
                        if self.reference:
                            urn += ":" + str(self.reference)
            self.__urn = urn
        return self.__urn

    def upTo(self, key):
        """ Returns the urn up to given level using URN Constants

        :param key: Identifier of the wished resource using URN constants
        :type key: int
        :returns: String representation of the partial URN requested
        :rtype: str

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    a.upTo(URN.TEXTGROUP) == "urn:cts:latinLit:phi1294"
        """
        middle = [
            component
            for component in [self.__parsed["textgroup"], self.__parsed["work"], self.__parsed["version"]]
            if component is not None
        ]

        if key == URN.COMPLETE:
            return self.__str__()
        elif key == URN.NAMESPACE:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"]])
        elif key == URN.TEXTGROUP and self.__parsed["textgroup"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                self.__parsed["textgroup"]
            ])
        elif key == URN.WORK and self.__parsed["work"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join([self.__parsed["textgroup"], self.__parsed["work"]])
            ])
        elif key == URN.VERSION and self.__parsed["version"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle)
            ])
        elif key == URN.NO_PASSAGE and self.__parsed["work"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle)
            ])
        elif key == URN.PASSAGE and self.__parsed["reference"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference)
            ])
        elif key == URN.PASSAGE_START and self.__parsed["reference"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference.start)
            ])
        elif key == URN.PASSAGE_END and self.__parsed["reference"] and self.reference.end is not None:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference.end)
            ])
        else:
            raise KeyError("Provided key is not recognized.")

    @staticmethod
    def model():
        return {
            "urn_namespace": None,
            "cts_namespace": None,
            "textgroup": None,
            "work": None,
            "version": None,
            "reference": None
        }

    def __parse(self, urn):
        """ Parse a URN

        :param urn: A URN:CTS
        :type urn: basestring
        :rtype: defaultdict.basestring
        :returns: Dictionary representation
        """
        parsed = URN.model()
        self.__urn = urn.split("#")[0]
        urn = self.__urn.split(":")
        if isinstance(urn, list) and len(urn) > 2:
            parsed["urn_namespace"] = urn[1]
            parsed["cts_namespace"] = urn[2]

            if len(urn) == 5:
                parsed["reference"] = Reference(urn[4])

            if len(urn) >= 4:
                urn = urn[3].split(".")
                if len(urn) >= 1:
                    parsed["textgroup"] = urn[0]
                if len(urn) >= 2:
                    parsed["work"] = urn[1]
                if len(urn) >= 3:
                    parsed["version"] = urn[2]
        else:
            raise ValueError("URN is empty")
        return parsed


class Citation(object):
    """ A citation object gives informations about the scheme 
        
    :param name: Name of the citation (e.g. "book")
    :type name: basestring
    :param xpath: Xpath of the citation (As described by CTS norm)
    :type xpath: basestring
    :param scope: Scope of the citation (As described by CTS norm)
    :type xpath: basestring
    :param refsDecl: refsDecl version
    :type refsDecl: basestring
    :param child: A citation
    :type child: Citation
    :ivar name: Name of the citation (e.g. "book")
    :type name: basestring
    :ivar xpath: Xpath of the citation (As described by CTS norm)
    :type xpath: basestring
    :ivar scope: Scope of the citation (As described by CTS norm)
    :type xpath: basestring
    :ivar refsDecl: refsDecl version
    :type refsDecl: basestring
    :ivar child: A citation
    :type child: Citation

    """

    def __init__(self, name=None, xpath=None, scope=None, refsDecl=None, child=None):
        """ Initialize a Citation object
        """
        self.__name = None
        self.__xpath = None
        self.__scope = None
        self.__refsDecl = None
        self.__child = None

        self.name = name
        self.scope = scope
        self.xpath = xpath
        self.refsDecl = refsDecl

        if child is not None:
            self.child = child

    @property
    def name(self): 
        """ Type of the citation represented
        
        :type: basestring
        :Example: Book, Chapter, Textpart, Section, Poem...
        """
        return self.__name

    @name.setter
    def name(self, val):
        self.__name = val

    @property
    def xpath(self):
        """ TextInventory xpath property of a citation (ie. identifier of the last element of the citation)
        
        :type: basestring
        :Example: //tei:l[@n="?"]
        """
        return self.__xpath

    @xpath.setter
    def xpath(self, val):
        if val is not None:
            self.__xpath = val
            self.__upRefsDecl()

    @property
    def scope(self):
        """ TextInventory scope property of a citation (ie. identifier of all element but the last of the citation)
        
        :type: basestring
        :Example: /tei:TEI/tei:text/tei:body/tei:div
        """
        return self.__scope
        
    @scope.setter
    def scope(self, val):
        if val is not None:
            self.__scope = val
            self.__upRefsDecl()

    @property
    def refsDecl(self):
        """ ResfDecl expression of the citation scheme 

        :type: basestring
        :Example: /tei:TEI/tei:text/tei:body/tei:div//tei:l[@n='$1']
        """
        return self.__refsDecl
        
    @refsDecl.setter
    def refsDecl(self, val):
        if val is not None:
            self.__refsDecl = val
            self.__upXpathScope()

    @property
    def child(self):
        """ Child of a citation

        :type: Citation or None
        :Example: Citation.name==poem would have a child Citation.name==line
        """
        return self.__child
        
    @child.setter
    def child(self, val):
        if isinstance(val, self.__class__):
            self.__child = val

    def __upXpathScope(self):
        """ Update xpath and scope property when refsDecl is updated
        
        """
        rd = self.__refsDecl
        matches = REFSDECL_SPLITTER.findall(rd)
        self.__scope = REFSDECL_REPLACER.sub("?", "".join(matches[0:-1]))
        self.__xpath = REFSDECL_REPLACER.sub("?", matches[-1])

    def __upRefsDecl(self):
        """ Update xpath and scope property when refsDecl is updated
        
        """
        if self.__scope is not None and self.__xpath is not None:
            xpath = self.__scope + self.__xpath
            i = xpath.find("?")
            ii = 1
            while i >= 0:
                xpath = xpath[:i] + "$" + str(ii) + xpath[i+1:]
                i = xpath.find("?")
                ii += 1
            self.__refsDecl = xpath

    def __iter__(self):
        """ Iteration method
        
        Loop over the citation childs

        :Example:
            >>>    c = Citation(name="line")
            >>>    b = Citation(name="poem", child=c)
            >>>    a = Citation(name="book", child=b)
            >>>    [e for e in a] == [a, b, c]
            
        """
        e = self
        while e is not None:
            yield e
            if hasattr(e, "child") and e.child is not None:
                e = e.child
            else:
                break

    def __len__(self):
       """ Length method

       :rtype: int
       :returns: Number of nested citations
       """
       return len([item for item in self])

    def fill(self, passage=None, xpath=None):
        """ Fill the xpath with given informations

        :param passage: Passage reference
        :type passage: Reference or list or None. Can be list of None and not None
        :param xpath: If set to True, will return the replaced self.xpath value and not the whole self.refsDecl
        :type xpath: Boolean
        :rtype: basestring
        :returns: Xpath to find the passage

        .. code-block:: python

            citation = Citation(name="line", scope="/TEI/text/body/div/div[@n=\"?\"]",xpath="//l[@n=\"?\"]")
            print(citation.fill(["1", None]))
            # /TEI/text/body/div/div[@n='1']//l[@n]
            print(citation.fill(None))
            # /TEI/text/body/div/div[@n]//l[@n]
            print(citation.fill(Reference("1.1"))
            # /TEI/text/body/div/div[@n='1']//l[@n='1']
            print(citation.fill("1", xpath=True)
            # //l[@n='1']


        """
        if xpath is True:  # Then passage is a string or None
            xpath = self.xpath

            if passage is None:
                replacement = r"\1"
            elif isinstance(passage, basestring):
                replacement = r"\1\2'" + passage + "'"

            return REFERENCE_REPLACER.sub(replacement, xpath)
        else:        
            if isinstance(passage, Reference):
                passage = passage.list or passage.start.list
            elif passage is None:
                return REFERENCE_REPLACER.sub(
                    r"\1",
                    self.refsDecl
                )
            passage = iter(passage)
            return REFERENCE_REPLACER.sub(
                lambda m: REF_REPLACER(m, passage),
                self.refsDecl
            )

    def __getstate__(self):
        """ Pickling method

        :return:
        """
        return copy(self.__dict__)

    def __setstate__(self, dic):
        self.__dict__ = dic
        return self


def REF_REPLACER(match, passage):
    """ Helper to replace xpath/scope/refsDecl on iteration with passage value

    :param match: A RegExp match
    :type match: re.SRE_MATCH
    :param passage: A list with subreference informations
    :type passage: iter

    :rtype: basestring
    :return: Replaced string
    """
    groups = match.groups()
    ref = next(passage)
    if ref is None:
        return groups[0]
    else:
        return "{1}='{0}'".format(ref, groups[0])





class DuplicateReference(SyntaxWarning):
    """ Error generated when a duplicate is found in Reference
    """

class RefsDeclError(Exception):
    """ Error issued when an the refsDecl does not succeed in xpath (no results)
    """
    pass


class InvalidSiblingRequest(Exception):
    """ This error is thrown when one attempts to get previous or next passage on a passage with a range of different
    depth, ex. : 1-2.25
    """
    pass


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.proto.inventory
   :synopsis: Prototypes for repository/inventory

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from MyCapytain.common.reference import URN, Reference, Citation
from MyCapytain.common.metadata import Metadata
from past.builtins import basestring
from collections import defaultdict
from copy import copy
from lxml import etree
from six import text_type as str


class Resource(object):
    """ Resource represents any resource from the inventory """
    def __init__(self, resource=None):
        """ Initiate a TextInventory resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        """
        self.metadata = Metadata()
        self.resource = None
        if resource is not None:
            self.setResource(resource)

    def __getitem__(self, key):
        """ Direct key access function for Text objects """
        if key == 0:
            return self
        elif isinstance(key, int) and 1 <= key <= len(self.parents):
            r = self.parents[key - 1]
            if isinstance(r, (list, tuple)):
                return r[0]
            else:
                return r
        elif isinstance(key, basestring):
            return self.__urnitem__(key)
        else:
            None

    def __eq__(self, other):
        if self is other:
            return True
        elif not isinstance(other, self.__class__):
            return False
        elif self.resource is None:
            # Not totally true
            return hasattr(self, "urn") and hasattr(other, "urn") and self.urn == other.urn
        return hasattr(self, "urn") and hasattr(other, "urn") and self.urn == other.urn and self.resource == other.resource

    def __str__(self):
        raise NotImplementedError()

    def export(self, format=None):
        raise NotImplementedError()

    def __urnitem__(self, key):
        urn = URN(key)

        if len(urn) <= 2:
            raise ValueError("Not valid urn")
        elif hasattr(self, "urn") and self.urn == urn:
            return self
        else:
            if hasattr(self, "urn"):
                i = len(self.urn)
            else:
                i = 2

            if isinstance(self, TextInventory):
                children = self.textgroups
            elif isinstance(self, TextGroup):
                children = self.works
            elif isinstance(self, Work):
                children = self.texts

            order = ["", "", URN.TEXTGROUP, URN.WORK, URN.VERSION]
            while i <= len(urn) - 1:
                children = children[urn.upTo(order[i])]
                if not hasattr(children, "urn") or str(children.urn) != urn.upTo(order[i]):
                    error = "Unrecognized urn at " + [
                        "URN namespace", "CTS Namespace", "URN Textgroup", "URN Work", "URN Version"
                    ][i]
                    raise ValueError(error)
                i += 1
            return children

    def setResource(self, resource):
        """ Set the object property resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :rtype: Any
        :returns: Input resource
        """
        self.resource = resource
        self.parse(resource)
        return self.resource

    def parse(self, resource):
        """ Parse the object resource 

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :rtype: List
        """
        raise NotImplementedError()

    def __getstate__(self, children=True):
        """ Pickling method to be called upon dumping object

        :return:
        """

        dic = copy(self.__dict__)
        if "xml" in dic:
            dic["xml"] = etree.tostring(dic["xml"], encoding=str)
        if "resource" in dic:
            del dic["resource"]
        """ The resource is unecessary in later than parsing state
        if "resource" in dic:
            dic["resource"] = str(dic["resource"])
        """
        return dic

    def __setstate__(self, dic):
        """

        :param dic:
        :return:
        """
        self.__dict__ = dic
        if "xml" in dic:
            self.xml = etree.fromstring(dic["xml"])
        return self


class Text(Resource):
    """ Represents a CTS Text
    """
    def __init__(self, resource=None, urn=None, parents=None, subtype="Edition"):
        """ Initiate a Work resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :param urn: Identifier of the Text
        :type urn: str
        """
        self.citation = None
        self.lang = None
        self.urn = None
        self.docname = None
        self.parents = list()
        self.subtype = subtype
        self.validate = None
        self.metadata = Metadata(keys=["label", "description", "namespaceMapping"])

        if urn is not None:
            self.urn = URN(urn)

        if parents is not None:
            self.parents = parents
            self.lang = self.parents[0].lang

        if resource is not None:
            self.setResource(resource)

    def translations(self, key=None):
        """ Get translations in given language

        :param key: Language ISO Code to filter on
        :return:
        """
        return self.parents[0].getLang(key)

    def editions(self):
        """ Get all editions of the texts

        :return: List of editions
        :rtype: [Text]
        """
        return [
                self.parents[0].texts[urn]
                for urn in self.parents[0].texts
                if self.parents[0].texts[urn].subtype == "Edition"
            ]


def Edition(resource=None, urn=None, parents=None):
    return Text(resource=resource, urn=urn, parents=parents, subtype="Edition")


def Translation(resource=None, urn=None, parents=None):
    return Text(resource=resource, urn=urn, parents=parents, subtype="Translation")


class Work(Resource):
    """ Represents a CTS Work
    """
    def __init__(self, resource=None, urn=None, parents=None):
        """ Initiate a Work resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :param urn: Identifier of the Work
        :type urn: str
        :param parents: List of parents for current object
        :type parents: Tuple.<TextInventory> 
        """
        self.lang = None
        self.urn = None
        self.texts = defaultdict(Text)
        self.parents = list()
        self.metadata = Metadata(keys=["title"])

        if urn is not None:
            self.urn = URN(urn)

        if parents is not None:
            self.parents = parents

        if resource is not None:
            self.setResource(resource)

    def getLang(self, key=None):
        """ Find a translation with given language

        :param key: Language to find
        :type key: basestring
        :rtype: [Text]
        :returns: List of availables translations
        """
        if key is not None:
            return [self.texts[urn] for urn in self.texts if self.texts[urn].subtype == "Translation" and self.texts[urn].lang == key]
        else:
            return [self.texts[urn] for urn in self.texts if self.texts[urn].subtype == "Translation"]


class TextGroup(Resource):
    """ Represents a CTS Textgroup
    """
    def __init__(self, resource=None, urn=None, parents=None):
        """ Initiate a TextGroup resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :param urn: Identifier of the TextGroup
        :type urn: str
        :param parents: List of parents for current object
        :type parents: Tuple.<TextInventory> 
        """
        self.urn = None
        self.works = defaultdict(Work)
        self.parents = list()
        self.metadata = Metadata(keys=["groupname"])

        if urn is not None:
            self.urn = URN(urn)

        if parents:
            self.parents = parents

        if resource is not None:
            self.setResource(resource)


class TextInventory(Resource):
    """ Represents a CTS Inventory file
    """
    def __init__(self, resource=None, id=None):
        """ Initiate a TextInventory resource

        :param resource: Resource representing the TextInventory 
        :type resource: Any
        :param id: Identifier of the TextInventory
        :type id: str
        """
        self.textgroups = defaultdict(TextGroup)
        self.id = id
        self.parents = list()
        if resource is not None:
            self.setResource(resource)

    def __len__(self):
        """

        :return: Number of texts available in the inventory
        """
        return len([
            text
                for tg in self.textgroups.values()
                for work in tg.works.values()
                for text in work.texts.values()
        ])



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml

Shared elements for TEI Citation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>
"""

import MyCapytain.common.reference
import MyCapytain.common.utils
import MyCapytain.resources.proto.text

from lxml.etree import _Element, tostring
from builtins import range, object
from six import text_type as str


def childOrNone(liste):
    if len(liste) > 0:
        return liste[-1]
    else:
        return None


class Citation(MyCapytain.common.reference.Citation):
    """ Implementation of Citation for TEI markup

    .. automethod:: __str__
    
    """
    def __str__(self):
        """ Returns a string refsDecl version of the object 

        :Example:
            >>>    a = Citation(name="book", xpath="/tei:TEI/tei:body/tei:text/tei:div", scope="/tei:div[@n=\"1\"]")
            >>>    str(a) == \"\"\"<tei:refsDecl n='book' xpath='/tei:TEI/tei:body/tei:text/tei:div' scope='/tei:div[@n=\"1\"]'>
            >>>         <tei:p>This Citation extracts Book from the text</tei:p>
            >>>    </tei:refsDecl>\"\"\"
        """
        if self.refsDecl is None:
            return ""

        label = ""
        if self.name is not None:
            label = self.name

        return "<tei:cRefPattern n=\"{label}\" matchPattern=\"{regexp}\" replacementPattern=\"#xpath({refsDecl})\">" \
               "<tei:p>This pointer pattern extracts {label}</tei:p></tei:cRefPattern>".format(
                   refsDecl=self.refsDecl,
                   label=label,
                   regexp="\.".join(["(\w+)" for i in range(0, self.refsDecl.count("$"))])
               )

    @staticmethod
    def ingest(resource, xpath=".//tei:cRefPattern"):
        """ Ingest a resource and store data in its instance

        :param resource: XML node cRefPattern or list of them in ASC hierarchy order (deepest to highest, eg. lines to poem to book)
        :type resource: lxml.etree._Element
        :param xpath: XPath to use to retrieve citation
        :type xpath: str

        :returns: A citation object
        :rtype: Citation
        """
        if len(resource) == 0 and isinstance(resource, list):
            return None
        elif isinstance(resource, list):
            resource = resource[0]
        elif not isinstance(resource, _Element):
            return None

        resource = resource.xpath(xpath, namespaces=MyCapytain.common.utils.NS)
        resources = []

        for x in range(0, len(resource)):
            resources.append(
                Citation(
                    name=resource[x].get("n"),
                    refsDecl=resource[x].get("replacementPattern")[7:-1],
                    child=childOrNone(resources)
                )
            )

        return resources[-1]


class Passage(MyCapytain.resources.proto.text.Passage):
    def __str__(self):
        """ Text based representation of the passage
    
        :rtype: basestring
        :returns: XML of the passage in string form 
        """
        return tostring(self.resource, encoding=str)

    def text(self, exclude=None):
        """ Text content of the passage

        :param exclude: Remove some nodes from text
        :type exclude: List
        :rtype: basestring
        :returns: Text of the xml node
        :Example:
            >>>    P = Passage(resource='<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>')
            >>>    P.text == "Ibis hello b ab excusso missus in astra sago. "
            >>>    P.text(exclude=["note"]) == "Ibis hello b ab excusso missus in astra sago. "


        """

        if exclude is None:
            exclude = ""
        else:
            exclude = "[{0}]".format(
                " and ".join(
                    "not(./ancestor-or-self::{0})".format(excluded)
                    for excluded in exclude
                )
            )

        return MyCapytain.common.utils.normalize(
            " ".join(
                [
                    element
                    for element
                    in self.resource.xpath(
                        ".//descendant-or-self::text()" + exclude,
                        namespaces=MyCapytain.common.utils.NS,
                        smart_strings=False
                    )
                ]
            )
        )

    @property
    def xml(self):
        """ XML Representation of the Passage

        :rtype: lxml.etree._Element
        :returns: XML element representing the passage
        """
        return self.resource



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain
   :platform: Unix, Windows
   :synopsis: Canonical Text Service and Canonical URN tool suite for Python

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

__name__ = "MyCapytain"
__version__ = "0.1.3"
__all__ = ["common", "endpoints", "resources"]



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.endpoints.ahab
   :synopsis: Ahab endpoint implementation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
import MyCapytain.endpoints.proto
import requests


class Ahab(MyCapytain.endpoints.proto.Ahab):

    """ 
    Basic integration of the proto.CTS abstraction
    """
    pass



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.metadata
   :synopsis: Metadata related objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from __future__ import unicode_literals


from collections import defaultdict, OrderedDict
from past.builtins import basestring
from builtins import range, object


class Metadatum(object):
    """ Metadatum object represent a single field of metadata

    :param name: Name of the field
    :type name: basestring
    :param children: List of tuples, where first element is the key, and second the value
    :type children: List

    :Example:
        >>>    a = Metadatum(name="label", [("lat", "Amores"), ("fre", "Les Amours")])
        >>>    print(a["lat"]) # == "Amores"

    .. automethod:: __getitem__
    .. automethod:: __setitem__
    .. automethod:: __iter__

    """

    def __init__(self, name, children=None):
        """ Initiate a Metadatum object
        """
        self.name = name
        self.children = OrderedDict()
        self.default = None

        if children is not None and isinstance(children, list):
            for tup in children:
                self[tup[0]] = tup[1]

    def __getitem__(self, key):
        """ Add an iterable access method

        Int typed key access to the *n* th registered key in the instance.
        If string based key does not exist, see for a default.

        :param key: Key of wished value
        :type key: basestring, tuple, int
        :returns: An element of children whose index is key

        :raises: KeyError if key is unknown (when using Int based key or when default is not set)

        :Example:
            >>>    a = Metadatum(name="label", [("lat", "Amores"), ("fre", "Les Amours")])
            >>>    print(a["lat"]) # Amores
            >>>    print(a[("lat", "fre")]) # Amores, Les Amours
            >>>    print(a[0]) # Amores
            >>>    print(a["dut"]) # Amores

        """
        if isinstance(key, int):
            items = list(self.children.keys())
            if key + 1 > len(items):
                raise KeyError()
            else:
                key = items[key]
        elif isinstance(key, tuple):
            return tuple([self[k] for k in key])

        if key not in self.children:
            if self.default is None:
                raise KeyError()
            else:
                return self.children[self.default]
        else:
            return self.children[key]

    def __setitem__(self, key, value):
        """ Register index key and value for the instance

        :param key: Index key(s) for the metadata
        :type key: basestring, list, tuple
        :param value: Values for the metadata
        :type value: basestring, list, tuple
        :returns: An element of children whose index is key

        :raises: `TypeError` if key is not basestring or tuple of basestring
        :raises: `ValueError` if key and value are list and are not the same size

        :Example:
            >>> a = Metadatum(name="label")

            >>> a["eng"] = "Illiad"
            >>> print(a["eng"]) # Illiad

            >>> a[("fre", "grc")] = ("Illiade", "Ἰλιάς")
            >>> print(a["fre"], a["grc"]) # Illiade, Ἰλιάς

            >>> a[("ger", "dut")] = "Iliade"
            >>> print(a["ger"], a["dut"]) # Iliade, Iliade
        """

        if isinstance(key, tuple):
            if not isinstance(value, (tuple, list)):
                value = [value]*len(key)
            if len(value) < len(key):
                raise ValueError("Less values than keys detected")
            for i in range(0, len(key)):
                self[key[i]] = value[i]
        elif not isinstance(key, basestring):
            raise TypeError(
                "Only basestring or tuple instances are accepted as key")
        else:
            self.children[key] = value
            if self.default is None:
                self.default = key

    def setDefault(self, key):
        """ Set a default key when a field does not exist

        :param key: An existing key of the instance
        :type key: basestring
        :returns: Default key
        :raises: `ValueError` If key is not registered

        :Example:
            >>>    a = Metadatum(name="label", [("lat", "Amores"), ("fre", "Les Amours")])
            >>>    a.setDefault("fre")
            >>>    print(a["eng"]) # == "Les Amours"

        """
        if key not in self.children:
            raise ValueError("Can not set a default to an unknown key")
        else:
            self.default = key
        return self.default

    def __iter__(self):
        """ Iter method of Metadatum

        :Example:
            >>> a = Metadata(name="label", [("lat", "Amores"), ("fre", "Les Amours")])
            >>> for key, value in a:
            >>>     print(key, value) # Print ("lat", "Amores") and then ("fre", "Les Amours")
        """
        i = 0
        for key in self.children:
            yield (key, self.children[key])
            i += 1

    def __len__(self):
        """ Get the length of the current Metadatum object

        :return: Number of variant of the metadatum
        :rtype: int

        :Example:
            >>> a = Metadata(name="label", [("lat", "Amores"), ("fre", "Les Amours")])
            >>> len(a) == 2
        """
        return len(self.children)

    def __getstate__(self):
        """ Pickling method

        :return:
        """
        return dict(
            name=self.name,
            langs=[(key, val) for key, val in self.children.items()],
            default=self.default
        )

    def __setstate__(self, dic):
        """ Unpickling method
        :param value:
        :return:
        """
        self.name = dic["name"]
        self.children = OrderedDict(dic["langs"])
        self.default = dic["default"]
        return self


class Metadata(object):
    """
        A metadatum aggregation object provided to centralize metadata

        :param key: A metadata field name
        :type key: List.<basestring>

        :ivar metadata: Dictionary of metadatum

        .. automethod:: __getitem__
        .. automethod:: __setitem__
        .. automethod:: __iter__
        .. automethod:: __len__
        .. automethod:: __add__
    """
    def __init__(self, keys=None):
        """ Initiate the object
        """
        self.metadata = defaultdict(Metadatum)
        self.__keys = []
        if keys is not None:
            for key in keys:
                self[key] = Metadatum(name=key)

    def __getitem__(self, key):
        """ Add a quick access system through getitem on the instance

        :param key: Index key representing a set of metadatum
        :type key: basestring, int, tuple
        :returns: An element of children whose index is key
        :raises: `KeyError` If key is not registered or recognized

        :Example:
            >>>    a = Metadata()
            >>>    m1 = Metadatum(name="title", [("lat", "Amores"), ("fre", "Les Amours")])
            >>>    m2 = Metadatum(name="author", [("lat", "Ovidius"), ("fre", "Ovide")])
            >>>    a[("title", "author")] = (m1, m2)

            >>>    a["title"] == m1
            >>>    a[0] == m1
            >>>    a[("title", "author")] == (m1, m2)


        """
        if isinstance(key, int):
            if key + 1 > len(self.__keys):
                raise KeyError()
            else:
                key = self.__keys[key]
        elif isinstance(key, tuple):
            return tuple([self[k] for k in key])

        if key not in self.metadata:
            raise KeyError()
        else:
            return self.metadata[key]

    def __setitem__(self, key, value):
        """ Set a new metadata field

        :param key: Name of metadatum field
        :type key: basestring, tuple
        :param value: Metadum dictionary
        :type value: Metadatum
        :returns: An element of children whose index is key

        :raises: `TypeError` if key is not basestring or tuple of basestring
        :raises: `ValueError` if key and value are list and are not the same size

        :Example:
            >>>    a = Metadata()

            >>>    a["title"] = Metadatum(name="title", [("lat", "Amores"), ("fre", "Les Amours")])
            >>>    print(a["title"]["lat"]) # Amores

            >>>    a[("title", "author")] = (
            >>>         Metadatum(name="title", [("lat", "Amores"), ("fre", "Les Amours")]),
            >>>         Metadatum(name="author", [("lat", "Ovidius"), ("fre", "Ovide")])
            >>>     )
            >>>    print(a["title"]["lat"], a["author"]["fre"]) # Amores, Ovide

        """
        if isinstance(key, tuple):
            if len(value) < len(key):
                raise ValueError("Less values than keys detected")
            for i in range(0, len(key)):
                self[key[i]] = value[i]
        elif not isinstance(key, basestring):
            raise TypeError(
                "Only basestring or tuple instances are accepted as key")
        else:
            if not isinstance(value, Metadatum) and isinstance(value, list):
                self.metadata[key] = Metadatum(key, value)
            elif isinstance(value, Metadatum):
                self.metadata[key] = value

            if key in self.metadata and key not in self.__keys:
                self.__keys.append(key)

    def __iter__(self):
        """ Iter method of Metadata

        :Example:
            >>> a = Metadata(("title", "desc", "author"))
            >>> for key, value in a:
            >>>     print(key, value) # Print ("title", "<Metadatum object>") then ("desc", "<Metadatum object>")...
        """
        i = 0
        for key in self.__keys:
            yield (key, self.metadata[key])
            i += 1

    def __add__(self, other):
        """ Merge Metadata objects together

        :param other: Metadata object to merge with the current one
        :type other: Metadata
        :returns: The merge result of both metadata object
        :rtype: Metadata

        :Example:
            >>> a = Metadata(name="label")
            >>> b = Metadata(name="title")
            >>> a + b == Metadata(name=["label", "title"])
        """
        from copy import deepcopy
        result = deepcopy(self)
        for metadata_key, metadatum in other:
            if metadata_key in self.__keys:
                for key, value in metadatum:
                    result[metadata_key][key] = value
            else:
                result[metadata_key] = metadatum
        return result

    def __len__(self):
        """ Returns the number of Metadatum registered in the object

        :rtype: int
        :returns: Number of metadatum objects

        :Example:
            >>>    a = Metadata(("title", "description", "author"))
            >>>    print(len(a)) # 3
        """
        return len(
            [
                k
                for k in self.__keys
                if isinstance(self.metadata[k], Metadatum)
            ]
        )

    def __getstate__(self):
        """ Pickling method

        :return:
        """
        return {
            key: getattr(value, "__getstate__")() for key, value in self.metadata.items()
        }

    def __setstate__(self, dic):
        """ Unpickling method
        :param dic: Dictionary with request valied
        :return:
        """
        self.metadata = defaultdict(Metadatum)

        self.__keys = []
        for key, value in dic.items():
            self.__keys.append(key)
            self.metadata[key] = getattr(Metadatum(name=value["name"]), "__setstate__")(value)
        return self



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from past.builtins import basestring
from six import text_type as str

import MyCapytain.resources.proto.text
import MyCapytain.resources.texts.tei
import MyCapytain.resources.inventory
import MyCapytain.endpoints.proto
import MyCapytain.common.metadata
import MyCapytain.common.utils
import MyCapytain.common.reference


class Text(MyCapytain.resources.proto.text.Text):
    """ Passage representing object prototype

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: An API endpoint
    :type resource: MyCapytain.endpoints.proto.CTS
    :param citation: Citation for children level
    :type citation: MyCapytain.resources.texts.tei.Citation
    :param id: Identifier of the subreference without URN informations
    :type id: List

    """

    DEFAULT_LANG = "eng"

    def __init__(self, urn, resource, citation=None, **kwargs):
        __doc__ = Text.__doc__
        super(Text, self).__init__(urn=urn, citation=citation, **kwargs)

        self._cRefPattern = None

        self.resource = resource

        if citation is not None:
            self.citation = citation

        if "metadata" in kwargs and isinstance(kwargs["metadata"], MyCapytain.common.metadata.Metadata):
            self.metadata = kwargs["metadata"]
        else:
            self.metadata = MyCapytain.common.metadata.Metadata([
                "groupname", "label", "title"
            ])

        self.passages = []

    def getValidReff(self, level=1, reference=None):
        """ Given a resource, Text will compute valid reffs

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param reference: Passage reference
        :type reference: Reference
        :rtype: list(str)
        :returns: List of levels
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        if level == -1:
            level = len(self.citation)

        xml = self.resource.getValidReff(
            level=level,
            urn=urn
        )
        xml = MyCapytain.common.utils.xmlparser(xml)
        self.__parse_request(xml.xpath("//ti:request", namespaces=MyCapytain.common.utils.NS)[0])

        for ref in xml.xpath(
            "//ti:reply//ti:urn/text()",
            namespaces=MyCapytain.common.utils.NS
        ):
            self.passages.append(ref)

        return self.passages

    def getPassage(self, reference=None):
        """ Retrieve a passage and store it in the object

        :param reference: Reference of the passage
        :type reference: MyCapytain.common.reference.Reference, or MyCapytain.common.reference.URN, or str or list(str)
        :rtype: Passage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if isinstance(reference, MyCapytain.common.reference.URN):
            urn = str(reference)
        elif isinstance(reference, MyCapytain.common.reference.Reference):
            urn = "{0}:{1}".format(self.urn, str(reference))
        elif isinstance(reference, str):
            urn = "{0}:{1}".format(self.urn, reference)
        elif isinstance(reference, list):
            urn = "{0}:{1}".format(self.urn, ".".join(reference))
        else:
            urn = str(self.urn)

        response = MyCapytain.common.utils.xmlparser(self.resource.getPassage(urn=urn))

        self.__parse_request(response.xpath("//ti:request", namespaces=MyCapytain.common.utils.NS)[0])
        return Passage(urn=urn, resource=response, parent=self)

    def getPassagePlus(self, reference=None):
        """ Retrieve a passage and informations around it and store it in the object

        :param reference: Reference of the passage
        :type reference: MyCapytain.common.reference.Reference or List of basestring
        :rtype: Passage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        response = MyCapytain.common.utils.xmlparser(self.resource.getPassagePlus(urn=urn))

        self.__parse_request(response.xpath("//ti:reply/ti:label", namespaces=MyCapytain.common.utils.NS)[0])
        return Passage(urn=urn, resource=response, parent=self)

    def __parse_request(self, xml):
        """

        :param xml:
        :return:

        .. TODO: Finish self.citation parsing
        """
        for node in xml.xpath(".//ti:groupname", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["groupname"][lang] = node.text

        for node in xml.xpath(".//ti:title", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["title"][lang] = node.text

        for node in xml.xpath(".//ti:label", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["label"][lang] = node.text

        # Need to code that p
        if self.citation is None:
            self.citation = MyCapytain.resources.inventory.Citation.ingest(
                xml,
                xpath=".//ti:citation[not(ancestor::ti:citation)]"
            )

    def getLabel(self):
        """ Retrieve metadata about the text

        :rtype: Metadata
        :returns: Dictionary with label informations
        """
        response = MyCapytain.common.utils.xmlparser(
            self.resource.getLabel(urn=str(self.urn))
        )

        self.__parse_request(response.xpath("//ti:reply/ti:label", namespaces=MyCapytain.common.utils.NS)[0])

        return self.metadata

    def getPrevNextUrn(self, reference):
        """ Get the previous URN of a reference of the text

        :param reference: Reference from which to find siblings
        :type reference: Reference
        :return: (Previous Passage Reference,Next Passage Reference)
        """
        _prev, _next = Passage.prevnext(
            self.resource.getPrevNextUrn(
                urn="{}:{}".format(
                    str(
                        MyCapytain.common.reference.URN(
                            str(self.urn)).upTo(MyCapytain.common.reference.URN.NO_PASSAGE)
                    ),
                    str(reference)
                )
            )
        )
        return _prev, _next

    def getFirstUrn(self, reference=None):
        """ Get the first children URN for a given resource

        :param reference: Reference from which to find child (If None, find first reference)
        :type reference: Reference, str
        :return: Children URN
        :rtype: URN
        """
        if reference:
            urn = "{}:{}".format(
                str(MyCapytain.common.reference.URN(str(self.urn)).upTo(MyCapytain.common.reference.URN.NO_PASSAGE)),
                str(reference)
            )
        else:
            urn = self.urn

        _first = Passage.firstUrn(
            self.resource.getFirstUrn(
                urn
            )
        )
        return _first

    @property
    def reffs(self):
        """ Get all valid reffs for every part of the Text

        :rtype: MyCapytain.resources.texts.tei.Citation
        """
        if self.citation is None:
            reffs = [self.getValidReff()]
            return reffs + [
                reff for reffs in [self.getValidReff(level=i) for i in range(2, len(self.citation) + 1)] for reff in reffs
            ]
        else:
            return [
                reff for reffs in [self.getValidReff(level=i) for i in range(1, len(self.citation) + 1)] for reff in reffs
            ]


class Passage(MyCapytain.resources.texts.tei.Passage):

    def __init__(self, urn, resource, *args, **kwargs):
        super(Passage, self).__init__(resource=resource, *args, **kwargs)

        self.urn = urn

        # Could be set during parsing
        self.__next = False
        self.__prev = False
        self.__first = False
        self.__last = False

        self.__parse()

    @property
    def first(self):
        """ Children passage

        :rtype: Passage
        :returns: Previous passage at same level
        """
        if self.__first is False:
            # Request the next urn
            self.__first = self.parent.getFirstUrn(reference=str(self.urn.reference))
        return self.__first

    @property
    def prev(self):
        """ Previous passage

        :rtype: Passage
        :returns: Previous passage at same level
        """
        if self.__prev is False:
            # Request the next urn
            self.__prev, self.__next = self.parent.getPrevNextUrn(reference=self.urn.reference)
        return self.__prev

    @property
    def next(self):
        """ Shortcut for getting the following passage

        :rtype: MyCapytain.common.reference.Reference
        :returns: Following passage reference
        """
        if self.__next is False:
            # Request the next urn
            self.__prev, self.__next = self.parent.getPrevNextUrn(reference=self.urn.reference)
        return self.__next

    def getNext(self):
        """ Shortcut for getting the following passage

        :rtype: Passage
        :returns: Following passage at same level
        """
        if self.next:
            return self.parent.getPassage(reference=self.next)

    def getPrev(self):
        """ Shortcut for getting the preceding passage

        :rtype: Passage
        :returns: Previous passage at same level
        """
        if self.prev:
            return self.parent.getPassage(reference=self.prev)

    def getFirst(self):
        """ Shortcut for getting the first child passage

        :rtype: Passage
        :returns: Previous passage at same level
        """
        if self.first:
            return self.parent.getPassage(reference=self.first)

    def __parse(self):
        """ Given self.resource, split informations from the CTS API

        :return: None
        """
        self.resource = self.resource.xpath("//ti:passage/tei:TEI", namespaces=MyCapytain.common.utils.NS)[0]

        self.__prev, self.__next = Passage.prevnext(self.resource)

    @staticmethod
    def prevnext(resource):
        """ Parse a resource to get the prev and next urn

        :param resource: XML Resource
        :type resource: etree._Element
        :return: Tuple representing previous and next urn
        :rtype: (URN, URN)
        """
        _prev, _next = False, False
        resource = MyCapytain.common.utils.xmlparser(resource)
        prevnext = resource.xpath("//ti:prevnext", namespaces=MyCapytain.common.utils.NS)

        if len(prevnext) > 0:
            _next, _prev = None, None
            prevnext = prevnext[0]
            _next_xpath = prevnext.xpath("ti:next/ti:urn/text()", namespaces=MyCapytain.common.utils.NS, smart_strings=False)
            _prev_xpath = prevnext.xpath("ti:prev/ti:urn/text()", namespaces=MyCapytain.common.utils.NS, smart_strings=False)

            if len(_next_xpath):
               _next = MyCapytain.common.reference.URN(_next_xpath[0])

            if len(_prev_xpath):
                _prev = MyCapytain.common.reference.URN(_prev_xpath[0])

        return _prev, _next

    @staticmethod
    def firstUrn(resource):
        """ Parse a resource to get the first URN

        :param resource: XML Resource
        :type resource: etree._Element
        :return: Tuple representing previous and next urn
        :rtype: (URN, URN)
        """
        _child = False
        resource = MyCapytain.common.utils.xmlparser(resource)
        urn = resource.xpath("//ti:reply/ti:urn/text()", namespaces=MyCapytain.common.utils.NS, magic_string=True)

        if len(urn) > 0:
            urn = str(urn[0])

            return MyCapytain.common.reference.URN(urn)




# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.endpoints
   :synopsis: Implementation and abstraction of CTS related endpoints

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml


Local files handler for CTS

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from collections import OrderedDict, defaultdict
from past.builtins import basestring
import warnings

from MyCapytain.errors import DuplicateReference, RefsDeclError
from MyCapytain.common.utils import xmlparser, NS, copyNode, passageLoop, normalizeXpath, normalize, \
    nested_set, nested_ordered_dictionary
from MyCapytain.common.reference import URN, Citation, Reference
from MyCapytain.resources.proto import text
from MyCapytain.errors import InvalidSiblingRequest
import MyCapytain.resources.texts.tei
from lxml import etree


class Text(text.Text):
    """ Implementation of CTS tools for local files

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: A resource
    :type resource: lxml.etree._Element
    :param citation: Highest Citation level
    :type citation: MyCapytain.common.reference.Citation
    :param autoreffs: Parse references on load (default : True)
    :type autoreffs: bool
    :ivar resource: lxml
    """

    def __init__(self, urn=None, citation=None, resource=None, autoreffs=False):
        super(Text, self).__init__(urn=urn, citation=citation)
        self._passages = Passage()
        self._orphan = defaultdict(Reference)  # Represents passage we got without asking for all. Storing convenience ?

        self._cRefPattern = MyCapytain.resources.texts.tei.Citation()
        self.xml = None

        if citation is not None:
            self.citation = citation

        if resource is not None:
            self.resource = resource
            self.xml = xmlparser(resource)

            self.__findCRefPattern(self.xml)

            if autoreffs is True:
                self.parse()

    def __findCRefPattern(self, xml):
        """ Find CRefPattern in the text and set object.citation
        :param xml: Xml Resource
        :return: None
        """
        self.citation = MyCapytain.resources.texts.tei.Citation.ingest(
            resource=xml.xpath("//tei:refsDecl[@n='CTS']", namespaces=MyCapytain.common.utils.NS),
            xpath=".//tei:cRefPattern"
        )

    def parse(self):
        """ Parse the object and generate the children
        """
        try:
            xml = self.xml.xpath(self.citation.scope, namespaces=NS)[0]
        except IndexError:
            msg = "Main citation scope does not result in any result ({0})".format(self.citation.scope)
            raise RefsDeclError(msg)
        except Exception as E:
            raise E

        self._passages = Passage(resource=xml, citation=self.citation, urn=self.urn, reference=None)

    @property
    def citation(self):
        """ Get the lowest cRefPattern in the hierarchy

        :rtype: Citation
        """
        return self._cRefPattern

    @citation.setter
    def citation(self, value):
        """ Set the cRefPattern

        :param value: Citation to be saved
        :type value: Citation
        :raises: TypeError when value is not a TEI Citation or a Citation
        """
        if isinstance(value, MyCapytain.resources.texts.tei.Citation):
            self._cRefPattern = value
        elif isinstance(value, Citation):
            # .. todo:: Should support conversion between Citation...
            self._cRefPattern = MyCapytain.resources.texts.tei.Citation(
                name=value.name,
                refsDecl=value.refsDecl,
                child=value.child
            )

    def nested_dict(self, exclude=None):
        """ Nested Dict Representation of the text passages

        :param exclude: Remove some nodes from text according to `MyCapytain.resources.texts.tei.Passage.text`
        :type exclude: List
        :rtype: dict
        :returns: Dictionary
        """
        reffs = self.getValidReff(level=len(self.citation))
        text = nested_ordered_dictionary()
        for reff in reffs:
            _r = reff.split(".")
            nested_set(text, _r, self.getPassage(_r, hypercontext=False).text(exclude=exclude))
        return text

    def getPassage(self, reference, hypercontext=True):
        """ Finds a passage in the current text

        :param reference: Identifier of the subreference / passages
        :type reference: list, Reference
        :param hypercontext: If set to true, retrieves nodes up to the given one, cleaning non required siblings.
        :type hypercontext: bool
        :rtype: Passage, ContextPassage
        :returns: Asked passage

        .. note :: As of MyCapytain 0.1.0, Text().getPassage() returns by default a ContextPassage, thus being able
            to handle range. This design change also means that the returned tree is way different that a classic
             Passage. To retrieve MyCapytain<=0.0.9 behaviour, use `hypercontext=False`.
        """
        if hypercontext is True:
            return self._getPassageContext(reference)

        if isinstance(reference, Reference):
            reference = reference.list or reference.start.list

        if self._passages.resource is None:
            self.parse()

        reference = [".".join(reference[:i]) for i in range(1, len(reference) + 1)]
        passages = [self._passages]
        while len(reference) > 0:
            passages = [passage for sublist in [p.get(reference[0]) for p in passages] for passage in sublist]
            reference.pop(0)

        return passages[0]

    def _getPassageContext(self, reference):
        """ Retrieves nodes up to the given one, cleaning non required siblings.

        :param reference: Identifier of the subreference / passages
        :type reference: list, reference
        :returns: Asked passage
        :rtype: ContextPassage
        """
        if isinstance(reference, list):
            start, end = reference, reference
            reference = Reference(".".join(reference))
        elif not reference.end:
            start, end = reference.start.list, reference.start.list
        else:
            start, end = reference.start.list, reference.end.list

        if len(start) > len(self.citation):
            raise ReferenceError("URN is deeper than citation scheme")

        citation_start = [citation for citation in self.citation][len(start)-1]
        citation_end = [citation for citation in self.citation][len(end)-1]

        start, end = citation_start.fill(passage=start), citation_end.fill(passage=end)

        nodes = etree._ElementTree()

        start, end = normalizeXpath(start.split("/")[2:]), normalizeXpath(end.split("/")[2:])

        root = copyNode(self.xml)
        nodes._setroot(root)
        root = passageLoop(self.xml, root, start, end)

        if self.urn:
            urn, reference = URN("{}:{}".format(self.urn, reference)), reference
        else:
            urn, reference = None, reference
        return ContextPassage(
            urn=urn,
            resource=root, parent=self, citation=self.citation
        )

    def getValidReff(self, level=1, reference=None):
        """ Retrieve valid passages directly

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: int
        :param reference: Passage Reference
        :type reference: Reference
        :returns: List of levels
        :rtype: list(basestring, str)

        :

        .. note:: GetValidReff works for now as a loop using Passage, subinstances of Text, to retrieve the valid
        informations. Maybe something is more powerfull ?

        """
        depth = 0
        xml = self.xml
        _range = False
        if reference:
            if isinstance(reference, Reference):
                if reference.end is None:
                    passages = [reference.list]
                else:
                    xml = self.getPassage(reference=reference)
                    a, b = reference.start.list, reference.end.list
                    passages = [[]]

            else:
                raise TypeError()

            depth = len(passages[0])
        else:
            passages = [[]]

        if level <= len(passages[0]) and reference is not None:
            level = len(passages[0]) + 1
        if level > len(self.citation):
            return []

        nodes = [None for i in range(depth, level)]

        citations = [citation for citation in self.citation]

        while len(nodes) >= 1:
            passages = [
                refs + [node.get("n")]
                for xpath_result, refs in [
                        (
                            xml.xpath(
                                citations[len(filling)-1].fill(filling),
                                namespaces=NS
                            ),
                            refs
                        )
                        for filling, refs in
                        [(refs + [None], refs) for refs in passages]
                    ]
                for node in xpath_result
            ]
            nodes.pop(0)

            if len(passages) == 0:
                msg = "Unknown reference {}".format(reference)
                raise KeyError(msg)

        passages = [".".join(passage) for passage in passages]
        duplicates = set([n for n in passages if passages.count(n) > 1])
        if len(duplicates) > 0:
            message = ", ".join(duplicates)
            warnings.warn(message, DuplicateReference)
        del duplicates

        return passages

    def text(self, exclude=None):
        """ Returns the text of the XML resource without the excluded nodes

        :param exclude: List of nodes
        :type exclude: list(str)
        :return: Text of the text without the text inside removed nodes
        :rtype: str

        .. example::
            `epigrammata.exclude(["tei:note"])` would remove all note nodes of the XML and print the text

        """
        return self._passages.text(exclude=exclude)


class Passage(MyCapytain.resources.texts.tei.Passage):
    """ Passage class for local texts which is fast but contains the minimum DOM.

        For design purposes, some people would prefer passage to be found quickly (Text indexing for example).
        Passage keeps only the node found through the xpath

        **Example** :  for a text with a citation scheme with following refsDecl :
        `/TEI/text/body/div[@type='edition']/div[@n='$1']/div[@n='$2']/l[@n='$3']` and a passage 1.1.1, this
        class will build an XML tree looking like the following

         .. code-block:: xml

            <l n='1'>Lorem ipsum</l>

    :param urn: A URN identifier
    :type urn: URN
    :param resource: A resource
    :type resource: etree._Element
    :param parent: Parent of the current passage
    :type parent: Passage
    :param citation: Citation for children level
    :type citation: Citation
    :param reference: Identifier of the subreference without URN information
    :type reference: Reference, List

    .. warning:: This passage system does not accept range
    """

    def __init__(self, urn=None, resource=None, parent=None, citation=None, reference=None):
        super(Passage, self).__init__(resource=resource, parent=parent)

        self.__next = False
        self.__prev = False

        self.citation = None
        if isinstance(citation, Citation):
            self.citation = citation

        self.__reference = Reference("")
        if urn:
            self.urn = urn
        if reference:
            self.reference = reference

        self.__children = OrderedDict()
        self.__parsed = False

    @property
    def reference(self):
        """ Id represents the passage subreference as a list of basestring

        :returns: Representation of the passage subreference as a list
        :rtype: Reference
        """
        return self.__reference

    @reference.setter
    def reference(self, value):
        """ Set up ID property

        :param value: Representation of the passage subreference as a list
        :type value: list, tuple, Reference

        .. note:: `Passage.id = [..]` will update automatically the URN property as well if correct
        """
        _value = None
        if isinstance(value, (list, tuple)):
            _value = Reference(".".join(value))
        elif isinstance(value, basestring):
            _value = Reference(value)
        elif isinstance(value, Reference):
            _value = value

        if _value and self.__reference != _value:
            self.__reference = _value
            if self._URN and len(self._URN):
                if len(value):
                    self._URN = URN("{}:{}".format(self._URN.upTo(URN.NO_PASSAGE), str(_value)))
                else:
                    self._URN = URN(self._URN["text"])

    @property
    def urn(self):
        """ URN Identifier of the object

        :rtype: URN
        """
        return self._URN

    @urn.setter
    def urn(self, value):
        """ Set the urn

        :param value: URN to be saved
        :type value: URN, basestring, str
        :raises: *TypeError* when the value is not URN compatible

        .. note:: `Passage.URN = ...` will update automatically the id property if Passage is set

        """
        a = self._URN

        if isinstance(value, basestring):
            value = URN(value)
        elif not isinstance(value, URN):
            raise TypeError()

        if str(a) != str(value):
            self._URN = value

            if value.reference and self.__reference != value.reference:
                self.__reference = value.reference
            elif not value.reference and self.__reference and len(self.__reference):
                self._URN = URN("{}:{}".format(str(value), str(self.__reference)))

    def get(self, key=None):
        """ Get a child or multiple children

        :param key: String identifying a passage
        :type key: basestring or int 
        :raises KeyError: When key identifies a child unknown to this passage
        :rtype: List.Passage
        :returns: List of passage identified by key. If key is None, returns all children
        
        .. note:: Call time depends on parsing status. If the passage was never parsed, then on first call citation is used to find children
        """
        if len(self.__children) == 0 and self.__parsed is False:
            self.__parse()

        if key is None:
            return [self.__children[key] for key in self.__children]
        elif isinstance(key, int):
            keys = list(self.__children.copy().keys())
            return [self.__children[keys[key]]]
        elif key not in self.__children:
            raise KeyError()
        else:
            return [self.__children[key]]

    def __parse(self):
        """ Private method for parsing children
        """
        if self.citation is None:
            self.__parsed = True
            return []

        elements = self.resource.xpath("."+self.citation.fill(passage=None, xpath=True), namespaces=NS)
        ids = [self.reference.list+[element.get("n")] for element in elements]
        ns = [".".join(_id) for _id in ids]

        # Checking for duplicates
        duplicates = set([n for n in ns if ns.count(n) > 1])
        if len(duplicates) > 0:
            message = ", ".join(duplicates)
            warnings.warn(message, DuplicateReference)

        for element, _id, n in zip(elements, ids, ns):
            self.__children[n] = Passage(
                resource=element,
                citation=self.citation.child,
                reference=_id,
                urn=self.urn,
                parent=self
            )
        self.__parsed = True

    @property
    def first(self):
        """ First child of current Passage 

        :returns: None if current Passage has no children,  first child passage if available
        :rtype: None, Passage
        """
        try:
            return self.get(0)[0]
        except:
            return None

    @property
    def last(self):
        """ Last child of current Passage 

        :returns: None if current Passage has no children, last child passage if available
        :rtype: None, Passage
        """
        try:
            return self.get(-1)[0]
        except :
            return None

    @property
    def children(self):
        """ Children of the passage

        :returns: Dictionary of chidren, where key are subreferences
        :rtype: OrderedDict
        """
        if len(self.__children) == 0 and self.__parsed is False:
            self.__parse()

        return self.__children
    
    @property
    def next(self):
        """ Next passage 

        :returns: Next passage at same level
        :rtype: Passage
        """ 
        if self.__next is False:

            if self.parent is None:  # When top of hierarchy is access, should return None
                self.__next = None
                return None

            keys = list(self.parent.children.copy().keys())
            current = keys.index(str(self.reference))
            if len(keys) - 1 > current:
                self.__next = self.parent.get(keys[current + 1])[0]
            else:
                n = self.parent.next
                if n is None:
                    self.__next = None
                else:
                    self.__next = n.first

        return self.__next

    @property
    def prev(self):
        """ Previous passage 

        :returns: Previous passage at same level
        :rtype: Passage
        """ 
        if self.__prev is False:

            if self.parent is None:  # When top of hierarchy is access, should return None
                self.__prev = None
                return None

            keys = list(self.parent.children.copy().keys())
            current = keys.index(str(self.reference))
            if current > 0:
                self.__prev = self.parent.get(keys[current - 1])[0]
            else:
                n = self.parent.prev
                if n is None:
                    self.__prev = None
                else:
                    self.__prev = n.last

        return self.__prev


class ContextPassage(Passage):
    """ Passage class for local texts which rebuilds the tree up to the passage.

        For design purposes, some people would prefer the output of GetPassage to be consistent. ContextPassage rebuilds
        the tree of the text up to the passage, keeping attributes of original nodes

        **Example** :  for a text with a citation scheme with following refsDecl :
        `/TEI/text/body/div[@type='edition']/div[@n='$1']/div[@n='$2']/l[@n='$3']` and a passage 1.1.1-1.2.3, this
        class will build an XML tree looking like the following

         .. code-block:: xml

            <TEI ...>
                <text ...>
                    <body ...>
                        <div type='edition' ...>
                            <div n='1' ...>

                                <div n='1' ...>
                                    <l n='1'>...</l>
                                    ...
                                </div>
                                <div n='2' ...>
                                    <l n='3'>...</l>
                                </div>
                            </div>
                        </div>
                    </body>
                </text>
            </TEI>

        :param urn: URN of the source text or of the passage
        :type urn: URN
        :param resource: Element representing the passage
        :type resource: etree._Element, Text
        :param parent: Text containing the passage
        :type parent: Text
        :param citation: Citation scheme of the text
        :type citation: Citation
        :param reference: Passage reference
        :type reference: Reference

        .. note::
            .prev, .next, .first and .last won't run on passage with a range made of two different level, such as
            1.1-1.2.3 or 1-a.b. Those will raise `InvalidSiblingRequest`

    """
    def __init__(self, urn=None, resource=None, parent=None, citation=None, reference=None):
        super(ContextPassage, self).__init__(urn=urn, reference=reference)

        if isinstance(resource, etree._Element):
            if urn:
                self.resource = Text(resource=resource, urn=urn.upTo(URN.NO_PASSAGE), citation=parent.citation)
            else:
                self.resource = Text(resource=resource, citation=parent.citation)
        else:
            self.resource = resource
        self.parent = parent
        self.citation = parent.citation
        self.__children = None

        self.depth = self.depth_2 = 1

        if self.reference.start:
            self.depth_2 = self.depth = len(self.reference.start)
        if self.reference and self.reference.end:
            self.depth_2 = len(self.reference.end)

        self.__prevnext = None  # For caching purpose

    def xpath(self, *args, **kwargs):
        """ Perform XPath on the passage XML

        :param args: Ordered arguments for etree._Element().xpath()
        :param kwargs: Named arguments
        :return: Result list
        :rtype: list(etree._Element)
        """
        if "smart_strings" not in kwargs:
            kwargs["smart_strings"] = False
        return self.resource.resource.xpath(*args, **kwargs)

    def tostring(self, *args, **kwargs):
        """ Transform the Passage in XML string

        :param args: Ordered arguments for etree.tostring() (except the first one)
        :param kwargs: Named arguments
        :return:
        """
        return etree.tostring(self.resource.resource, *args, **kwargs)

    def __str__(self):
        """ Text based representation of the passage

        :returns: XML of the passage in string form
        :rtype: basestring
        """
        return self.tostring(encoding=str)

    @property
    def first(self):
        """ First child of current Passage

        :returns: None if current Passage has no children,  first child passage if available
        :rtype: None, Reference
        """
        if self.depth >= len(self.citation):
            return None
        else:
            return self.children[0]

    @property
    def last(self):
        """ Last child of current Pass

        :returns: None if current Passage has no children, last child passage if available
        :rtype: None, Reference
        """
        if self.depth >= len(self.citation):
            return None
        else:
            return self.children[-1]

    @property
    def children(self):
        """ Children of the passage

        :rtype: None, Reference
        :returns: Dictionary of chidren, where key are subreferences
        """
        self.__raiseDepth()
        if self.depth >= len(self.citation):
            return []
        elif self.__children and len(self.__children):
            return self.__children
        else:
            self.__children = self.resource.getValidReff(level=self.depth+1)
            return self.__children

    @property
    def next(self):
        """ Next passage

        :returns: Next passage at same level
        :rtype: None, Reference
        """
        return self.__getSiblings(direction=1)

    @property
    def prev(self):
        """ Get the Previous passage reference

        :returns: Previous passage reference at the same level
        :rtype: None, Reference
        """
        return self.__getSiblings(direction=0)

    def text(self, exclude=None):
        """ Text content of the passage

        :param exclude: Remove some nodes from text
        :type exclude: List
        :rtype: basestring
        :returns: Text of the xml node
        :Example:
            >>>    P = Passage(resource='<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>')
            >>>    P.text == "Ibis hello b ab excusso missus in astra sago. "
            >>>    P.text(exclude=["note"]) == "Ibis hello b ab excusso missus in astra sago. "


        """

        if exclude is None:
            exclude = ""
        else:
            exclude = "[{0}]".format(
                " and ".join(
                    "not(./ancestor-or-self::{0})".format(excluded)
                    for excluded in exclude
                )
            )

        return normalize(
            " ".join(
                [
                    element
                    for element
                    in self.xpath(
                        ".//descendant-or-self::text()" + exclude,
                        namespaces=NS
                    )
                ]
            )
        )

    def __raiseDepth(self):
        """ Simple check that raises an exception if the passage cannot run first, last, next or prev

        See object notes

        :raise: InvalidSiblingRequest
        """
        if self.depth != self.depth_2:
            raise InvalidSiblingRequest()

    def __getSiblings(self, direction=1):
        """

        :return: Reference
        """
        self.__raiseDepth()

        if self.__prevnext:
            return self.__prevnext[direction]

        document_references = list(map(lambda x: str(x), self.parent.getValidReff(level=self.depth)))
        range_length = len(self.resource.getValidReff(level=self.depth))

        if self.reference.end:
            start, end = str(self.reference.start), str(self.reference.end)
        else:
            start = end = str(self.reference.start)

        start = document_references.index(start)
        end = document_references.index(end)

        _prev, _next = None, None

        if start == 0:
            # If the passage is already at the beginning
            _prev = None
        elif start - range_length < 0:
            if start == end:
                _prev = Reference(document_references[0])
            else:
                _prev = Reference(
                    "{}-{}".format(document_references[0], document_references[start-1])
                )
        else:
            if start == end:
                _prev = Reference(document_references[start-1])
            else:
                _prev = Reference(
                    "{}-{}".format(document_references[start-range_length], document_references[start-1])
                )

        if start + 1 == len(document_references) or end + 1 == len(document_references):
            # If the passage is already at the end
            _next = None
        elif end + range_length > len(document_references):
            if start == end:
                _next = Reference(document_references[-1])
            else:
                _next = Reference(
                    "{}-{}".format(document_references[end +1], document_references[-1])
                )
        else:
            if start == end:
                _next = Reference(document_references[end +1])
            else:
                _next = Reference(
                    "{}-{}".format(document_references[end + 1], document_references[end + range_length])
                )

        self.__prevnext = (_prev, _next)
        return self.__prevnext[direction]




# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: Common useful tools and constants

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from __future__ import unicode_literals
from functools import reduce

from collections import OrderedDict
from lxml import etree
from io import IOBase, StringIO
from past.builtins import basestring
import re
from copy import copy

__strip = re.compile("([ ]{2,})+")


def normalize(string):
    """ Remove double-or-more spaces in a string

    :param string: A string to change
    :type string: basestring
    :rtype: Basestring
    :returns: Clean string
    """
    return __strip.sub(" ", string)

#: Dictionary of namespace that can be useful
NS = {
    "tei": "http://www.tei-c.org/ns/1.0",
    "ahab": "http://localhost.local",
    "ti": "http://chs.harvard.edu/xmlns/cts",
    "xml": "http://www.w3.org/XML/1998/namespace"
}


def xmlparser(xml):
    """ Parse xml 

    :param xml: XML element
    :type xml: basestring, lxml.etree._Element
    :rtype: lxml.etree._Element
    :returns: An element object
    :raises: TypeError if element is not in accepted type

    """
    doclose = None
    if isinstance(xml, etree._Element):
        return xml
    elif isinstance(xml, IOBase):
        pass
    elif isinstance(xml, basestring):
        xml = StringIO(xml)
        doclose = True
    else:
        raise TypeError("Unsupported type of resource")
    parsed = etree.parse(xml).getroot()
    if doclose:
        xml.close()
    return parsed


def formatXpath(xpath):
    """

    :param xpath:
    :return:
    """
    if len(xpath) > 1:
        current, queue = xpath[0], xpath[1:]
        current = "./{}[./{}]".format(
            current,
            "/".join(queue)
        )
    else:
        current, queue = "./{}".format(xpath[0]), []

    return current, queue


def performXpath(parent, xpath):
    """

    :param parent:
    :param xpath:
    :return: (Result, Loop Indicator)
    """
    loop = False
    if xpath.startswith(".//"):
        result = parent.xpath(
            xpath.replace(".//", "./"),
            namespaces=NS
        )
        if len(result) == 0:
            result = parent.xpath(
                "*[{}]".format(xpath),
                namespaces=NS
            )
            loop = True
    else:
        result = parent.xpath(
            xpath,
            namespaces=NS
        )
    return result[0], loop


def copyNode(node, children=False, parent=False):
    """

    :param node:
    :param children:
    :param parent:
    :return:
    """
    if parent is not False:
        element = etree.SubElement(
            parent,
            node.tag,
            attrib=node.attrib,
            nsmap={None: "http://www.tei-c.org/ns/1.0"}
        )
    else:
        element = etree.Element(
            node.tag,
            attrib=node.attrib,
            nsmap={None: "http://www.tei-c.org/ns/1.0"}
        )
    if children:
        element.text = node.text
        for child in node:
            element.append(copy(child))
    return element


def normalizeXpath(xpath):
    """ Normalize XPATH splitted around slashes

    :param xpath: List of xpath elements
    :type xpath: [str]
    :return: List of refined xpath
    :rtype: [str]
    """
    new_xpath = []
    for x in range(0, len(xpath)):
        if x > 0 and len(xpath[x-1]) == 0:
            new_xpath.append("/"+xpath[x])
        elif len(xpath[x]) > 0:
            new_xpath.append(xpath[x])
    return new_xpath


def passageLoop(parent, new_tree, xpath1, xpath2=None, preceding_siblings=False, following_siblings=False):
    """

    :param parent: Parent on which to perform xpath
    :param new_tree: Parent on which to add nodes
    :param xpath: List of xpath elements
    :type xpath: [str]
    :return:
    """

    current_1, queue_1 = formatXpath(xpath1)
    if xpath2 is None:  # In case we need what is following or preceding our node
        result_1, loop = performXpath(parent, current_1)
        if loop is True:
            queue_1 = xpath1
        siblings = list(parent)
        index_1 = siblings.index(result_1)
        children = len(queue_1) == 0

        # We fill the gaps using the list option of LXML
        if preceding_siblings:
            [
                copyNode(child, parent=new_tree, children=True)
                for child in siblings
                if index_1 > siblings.index(child)
            ]
            child = copyNode(result_1, children=children, parent=new_tree)
        elif following_siblings:
            child = copyNode(result_1, children=children, parent=new_tree)
            [
                copyNode(child, parent=new_tree, children=True)
                for child in siblings
                if index_1 < siblings.index(child)
            ]

        if not children:
            child = passageLoop(
                result_1,
                child,
                queue_1,
                None,
                preceding_siblings=preceding_siblings,
                following_siblings=following_siblings
            )
    else:

        result_1, loop = performXpath(parent, current_1)
        if loop is True:
            queue_1 = xpath1
            if xpath2 == xpath1:
                current_2, queue_2 = current_1, queue_1
            else:
                current_2, queue_2 = formatXpath(xpath2)
        else:
            current_2, queue_2 = formatXpath(xpath2)

        if xpath1 != xpath2:
            result_2, loop = performXpath(parent, current_2)
            if loop is True:
                queue_2 = xpath2
        else:
            result_2 = result_1

        if result_1 == result_2:
            children = len(queue_1) == 0
            child = copyNode(result_1, children=children, parent=new_tree)
            if not children:
                child = passageLoop(
                    result_1,
                    child,
                    queue_1,
                    queue_2
                )
        else:
            children = list(parent)
            index_1 = children.index(result_1)
            index_2 = children.index(result_2)
            # Appends the starting passage
            children_1 = len(queue_1) == 0
            child_1 = copyNode(result_1, children=children_1, parent=new_tree)
            if not children_1:
                passageLoop(result_1, child_1, queue_1, None, following_siblings=True)
            # Appends what's in between
            nodes = [
                copyNode(child, parent=new_tree, children=True)
                for child in children
                if index_1 < children.index(child) < index_2
            ]
            # Appends the Ending passage
            children_2 = len(queue_2) == 0
            child_2 = copyNode(result_2, children=children_2, parent=new_tree)

            if not children_2:
                passageLoop(result_2, child_2, queue_2, None, preceding_siblings=True)

    return new_tree


class OrderedDefaultDict(OrderedDict):
    def __init__(self, default_factory=None, *args, **kwargs):
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        val = self[key] = self.default_factory()
        return val


def nested_ordered_dictionary():
    return OrderedDefaultDict(nested_ordered_dictionary)


def nested_get(dictionary, keys):
    """ Get value in dictionary for dictionary[keys[0]][keys[1]][keys[..n]]

    :param dictionary: An input dictionary
    :param keys: Keys where to store data
    :return:
    """
    return reduce(lambda d, k: d[k], keys, dictionary)


def nested_set(dictionary,  keys, value):
    """ Set value in dictionary for dictionary[keys[0]][keys[1]][keys[..n]]

    :param dictionary: An input dictionary
    :param keys: Keys where to store data
    :param value: Value to set at keys** target
    :return: None
    """
    nested_get(dictionary, keys[:-1])[keys[-1]] = value



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml
   :synopsis: XML based Text and repository

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from __future__ import unicode_literals

from MyCapytain.resources.proto import inventory, text
from MyCapytain.common.reference import Citation as CitationPrototype, URN
from MyCapytain.common.utils import xmlparser, NS
import re
from six import text_type as str
import collections


class Citation(CitationPrototype):
    """ Citation XML implementation for TextInventory

    """

    escape = re.compile('(")')

    def __str__(self):
        """ Returns a string text inventory version of the object 

        :Example:
            >>>    a = Citation(name="book", xpath="/tei:TEI/tei:body/tei:text/tei:div", scope="/tei:div[@n=\"1\"]")
            >>>    str(a) == <ti:citation label='book' xpath='/tei:TEI/tei:body/tei:text/tei:div' scope='/tei:div[@n=\"1\"]'>...</ti:citation>
        """
        if self.xpath is None and self.scope is None and self.refsDecl is None:
            return ""

        child = ""
        if isinstance(self.child, Citation):
            child = str(self.child)

        label = ""
        if self.name is not None:
            label = self.name

        return """<ti:citation label="{label}" xpath="{xpath}" scope="{scope}">{child}</ti:citation>""".format(
            child=child,
            xpath=re.sub(Citation.escape, "'", self.xpath),
            scope=re.sub(Citation.escape, "'", self.scope),
            label=label
        )

    @staticmethod
    def ingest(resource, element=None, xpath="ti:citation"):
        """ Ingest xml to create a citation

        :param xml: XML on which to do xpath
        :param element: Element where the citation should be stored
        :param xpath: XPath to use to retrieve citation

        :return: Citation
        """
        # Reuse of of find citation
        results = resource.xpath(xpath, namespaces=NS)
        if len(results) > 0:
            citation = Citation(
                name=results[0].get("label"),
                xpath=results[0].get("xpath"),
                scope=results[0].get("scope")
            )

            if isinstance(element, Citation):
                element.child = citation
                Citation.ingest(
                    resource=results[0],
                    element=element.child
                )
            else:
                element = citation
                Citation.ingest(
                    resource=results[0],
                    element=element
                )

            return citation

        return None


def xpathDict(xml, xpath, children, parents, **kwargs):
    """ Returns a default Dict given certain informations
        

    :param xml: An xml tree
    :type xml: etree
    :param xpath: XPath to find children
    :type basestring:
    :param children: Object identifying children
    :type children: inventory.Resource
    :param parents: Tuple of parents
    :type parents: tuple.<inventory.Resource>
    :rtype: collections.defaultdict.<basestring, inventory.Resource>
    :returns: Dictionary of children
    """
    return collections.defaultdict(children, **dict(
        (
            child.get("urn"),
            children(
                resource=child,
                urn=child.get("urn"),
                parents=parents,
                **kwargs
            )
        ) for child in xml.xpath(xpath, namespaces=NS))
    )


class Text(inventory.Text):
    """ Represents a CTS Text
        
        ..automethod:: __str__
    """
    def __init__(self, **kwargs):
        super(Text, self).__init__(**kwargs)

    def __str__(self):
        """ Print the xml of the text
        
        :rtype: basestring
        :returns: XML representation of the text
        """
        strings = []
        tag_start = "edition"
        tag_end = tag_start
        if self.subtype == "Translation":
            tag_start = "translation"
            tag_end = "translation"
            if self.lang:
                tag_start = tag_start + " xml:lang='" + self.lang + "'"

        if self.urn is not None:
            strings.append(
                "<ti:{0} urn='{1}' workUrn='{2}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(
                    tag_start,
                    self.urn,
                    self.urn.upTo(URN.WORK)
                )
            )
        else:
            if len(self.parents) > 0 and hasattr(self.parents[0], "urn") is True:
                strings.append(
                    "<ti:{0} workUrn='{1}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(
                        tag_start,
                        self.parents[0].urn
                    )
                )
            else:
                strings.append(
                    "<ti:{0} xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(
                        tag_start
                    )
                )

        namespaces = []
        for tag, metadatum in self.metadata:
            if tag == "namespaceMapping":
                for abbr, ns in metadatum:
                    namespaces.append(
                        '<ti:namespaceMapping abbreviation=\'{0}\' nsURI=\'{1}\'/>'.format(
                            abbr,
                            ns
                        )
                    )
            else:
                for lang, value in metadatum:
                    strings.append(
                        "<ti:{tag} xml:lang='{lang}'>{value}</ti:{tag}>".format(
                            tag=tag,
                            lang=lang,
                            value=value
                        )
                    )

        # Maybe should have an online object...
        docname = ""
        if self.docname is not None:
            docname = ' docname=\'{0}\''.format(self.docname)

        strings.append("<ti:online{0}>".format(docname))

        if self.validate is not None:
            strings.append('<ti:validate schema=\'{0}\'/>'.format(self.validate))

        if len(namespaces) > 0:
            strings.append("".join(namespaces))

        if self.citation is not None:
            strings.append("<ti:citationMapping>")
            strings.append(str(self.citation))
            strings.append("</ti:citationMapping>")

        strings.append("</ti:online>")

        strings.append("</ti:{0}>".format(tag_end))
        return "".join(strings)

    def export(self, output="xml", **kwargs):
        """ Create a {format} version of the Work
        
        :param output: Format to be chosen (Only XML for now)
        :type output: basestring, citation
        :rtype: lxml.etree._Element
        :returns: XML representation of the object
        """
        if output == "xml":
            return xmlparser(str(self))
        elif issubclass(output, text.Text):
            complete_metadata = self.metadata
            for parent in self.parents:
                if isinstance(parent, inventory.Resource) and hasattr(parent, "metadata"):
                    complete_metadata = complete_metadata + parent.metadata
            return output(urn=self.urn, citation=self.citation, metadata=complete_metadata, **kwargs)

    def __findCitations(self, xml, xpath="ti:citation"):
        """ Find citation in current xml. Used as a loop for self.xmlparser()
        
        :param xml: Xml resource to be parsed
        :param xpath: Xpath to use to retrieve the xml node
        """
        self.citation = Citation.ingest(xml, self.citation, xpath)

    def parse(self, resource):
        """ Parse a resource to feed the object
        
        :param resource: An xml representation object
        :type resource: basestring or lxml.etree._Element
        :returns: None
        """
        self.xml = xmlparser(resource)

        if self.subtype == "Translation":
            lang = self.xml.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lang is not None:
                self.lang = lang

        for child in self.xml.xpath("ti:description", namespaces=NS):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                self.metadata["description"][lg] = child.text

        for child in self.xml.xpath("ti:label", namespaces=NS):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                self.metadata["label"][lg] = child.text

        self.__findCitations(
            xml=self.xml,
            xpath="ti:online/ti:citationMapping/ti:citation"
        )

        online = self.xml.xpath("ti:online", namespaces=NS)
        if len(online) > 0:
            online = online[0]
            self.docname = online.get("docname")
            for validate in online.xpath("ti:validate", namespaces=NS):
                self.validate = validate.get("schema")
            for namespaceMapping in online.xpath("ti:namespaceMapping", namespaces=NS):
                self.metadata["namespaceMapping"][namespaceMapping.get("abbreviation")] = namespaceMapping.get("nsURI")

        return None


def Edition(resource=None, urn=None, parents=None):
    """ Create an edition subtyped Text object 
    """
    return Text(resource=resource, urn=urn, parents=parents, subtype="Edition")


def Translation(resource=None, urn=None, parents=None):
    """ Create a translation subtyped Text object 
    """
    return Text(resource=resource, urn=urn, parents=parents, subtype="Translation")


class Work(inventory.Work):

    """ Represents a CTS Textgroup in XML

        ..automethod:: __str__
    """

    def __init__(self, **kwargs):

        super(Work, self).__init__(**kwargs)

    def __str__(self):
        """ Print the xml of the work
        
        :rtype: basestring
        :returns: XML representation of the work
        """
        strings = []
        __lang = ""
        if self.lang:
            __lang = """xml:lang="{}" """.format(self.lang)

        if self.urn is not None:
            strings.append(
                "<ti:work {2}urn='{0}' groupUrn='{1}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(
                    self.urn, self.urn.upTo(URN.TEXTGROUP), __lang)
            )
        else:
            if len(self.parents) > 0 and hasattr(self.parents[0], "urn") is True:
                strings.append("<ti:work {1}groupUrn='{0}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(
                    self.parents[0].urn, __lang)
                )
            else:
                strings.append("<ti:work {}xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(__lang))
        for tag, metadatum in self.metadata:
            for lang, value in metadatum:
                strings.append("<ti:{tag} xml:lang='{lang}'>{value}</ti:{tag}>".format(tag=tag, lang=lang, value=value))

        # Dev trick : For tests, we need to have always the same order....
        keys = sorted([urn for urn in self.texts])
        for urn in keys:
            strings.append(str(self.texts[urn]))

        strings.append("</ti:work>")
        return "".join(strings)

    def export(self, output="xml"):
        """ Create a {format} version of the Work
        
        :param output: Format to be chosen (Only XML for now)
        :type output: basestring
        :rtype: lxml.etree._Element
        :returns: XML representation of the object
        """
        return xmlparser(str(self))

    def parse(self, resource):
        """ Parse a resource 

        :param resource: Element rerpresenting a work
        :param type: basestring, etree._Element
        """
        self.xml = xmlparser(resource)

        lang = self.xml.get("{http://www.w3.org/XML/1998/namespace}lang")
        if lang is not None:
            self.lang = lang

        for child in self.xml.xpath("ti:title", namespaces=NS):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                self.metadata["title"][lg] = child.text

        self.__editions = xpathDict(
            xml=self.xml,
            xpath='ti:edition',
            children=Edition,
            parents=[self] + self.parents
        )
        self.__translations = xpathDict(
            xml=self.xml,
            xpath='ti:translation',
            children=Translation,
            parents=[self] + self.parents
        )

        self.texts = collections.defaultdict(Text)
        for urn in self.__editions:
            self.texts[urn] = self.__editions[urn]
        for urn in self.__translations:
            self.texts[urn] = self.__translations[urn]

        return self.texts


class TextGroup(inventory.TextGroup):

    """ Represents a CTS Textgroup in XML
        
        .. automethod:: __str__
    """

    def __init__(self, **kwargs):
        super(TextGroup, self).__init__(**kwargs)

    def __str__(self):
        """ Print the xml of the text group
        
        :rtype: basestring
        :returns: XML representation of the textgroup
        """
        strings = []
        if self.urn is not None:
            strings.append("<ti:textgroup urn='{0}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(self.urn))
        else:
            strings.append("<ti:textgroup xmlns:ti='http://chs.harvard.edu/xmlns/cts'>")

        for tag, metadatum in self.metadata:
            for lang, value in metadatum:
                strings.append("<ti:{tag} xml:lang='{lang}'>{value}</ti:{tag}>".format(tag=tag, lang=lang, value=value))

        for urn in self.works:
            strings.append(str(self.works[urn]))

        strings.append("</ti:textgroup>")
        return "".join(strings)

    def export(self, output="xml"):
        """ Create a {format} version of the TextInventory
        
        :param output: Format to be chosen (Only XML for now)
        :type output: basestring
        :rtype: lxml.etree._Element
        :returns: XML representation of the object
        """
        return xmlparser(str(self))

    def parse(self, resource):
        """ Parse a resource 

        :param resource: Element rerpresenting the textgroup
        :param type: basestring, etree._Element
        """
        self.xml = xmlparser(resource)

        for child in self.xml.xpath("ti:groupname", namespaces=NS):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                self.metadata["groupname"][lg] = child.text

        self.works = xpathDict(
            xml=self.xml,
            xpath='ti:work',
            children=Work,
            parents=[self] + self.parents
        )
        return self.works


class TextInventory(inventory.TextInventory):

    """ Represents a CTS Inventory file
        
    .. automethod:: __str__
    """

    def __init__(self, **kwargs):
        super(TextInventory, self).__init__(**kwargs)

    def __str__(self):
        """ Print the xml of the textinventory
        
        :rtype: basestring
        :returns: XML representation of the textinventory
        """
        strings = []
        if self.id is not None:
            strings.append("<ti:TextInventory tiid='{0}' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>".format(self.id))
        else:
            strings.append("<ti:TextInventory xmlns:ti='http://chs.harvard.edu/xmlns/cts'>")

        for urn in self.textgroups:
            strings.append(str(self.textgroups[urn]))
        strings.append("</ti:TextInventory>")
        return "".join(strings)

    def export(self, output="xml"):
        """ Create a {output} version of the TextInventory
        
        :param output: output to be chosen (Only XML for now)
        :type output: basestring
        :rtype: lxml.etree._Element
        :returns: XML representation of the object
        """
        return xmlparser(str(self))

    def parse(self, resource):
        """ Parse a resource 

        :param resource: Element representing the text inventory
        :param type: basestring, etree._Element
        """
        self.xml = xmlparser(resource)

        self.textgroups = xpathDict(
            xml=self.xml,
            xpath='//ti:textgroup',
            children=TextGroup,
            parents=[self]
        )
        return self.textgroups



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.proto.text
   :synopsis: Prototypes for Text

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from past.builtins import basestring

import MyCapytain.common.reference
import MyCapytain.common.metadata
from collections import namedtuple

PassagePlus = namedtuple("PassagePlus", ["passage", "prev", "next"])


class Resource(object):
    """ Initiate a Resource object
    
    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: A resource
    :type resource: Any
    """

    def __init__(self, urn=None, resource=None):
        self.resource = None
        self._URN = None

        if urn is not None:
            self._URN = urn

        if resource is not None:
            self.resource = resource

    @property
    def urn(self):
        """ URN Identifier of the object

        :rtype: MyCapytain.common.reference.URN

        """
        return self._URN
    
    @urn.setter
    def urn(self, value):
        """ Set the urn

        :param value: URN to be saved
        :type value:  MyCapytain.common.reference.URN
        :raises: *TypeError* when the value is not URN compatible

        """
        if isinstance(value, basestring):
            value = MyCapytain.common.reference.URN(value)
        elif not isinstance(value, MyCapytain.common.reference.URN):
            raise TypeError()
        self._URN = value


class Passage(Resource):
    """ Passage representing object prototype
    
    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: A resource
    :type resource: lxml.etree._Element
    :param parent: Parent of the current passage
    :type parent: MyCapytain.resources.texts.tei.Passage
    :param citation: Citation for children level
    :type citation: MyCapytain.resources.texts.tei.Citation
    :param id: Identifier of the subreference without URN informations
    :type id: List
    
    """

    def __init__(self, parent=None, **kwargs):
        super(Passage, self).__init__(**kwargs)
        self.parent = None
        if parent is not None and isinstance(parent, Passage):
            self.parent = parent
        elif isinstance(parent, Text):
            self.parent = parent

    @property
    def prev(self):
        """ Previous passage 

        :rtype: Passage
        :returns: Previous passage at same level
        """ 
        raise NotImplementedError()

    @property
    def next(self):
        """ Following passage 

        :rtype: Passage
        :returns: Following passage at same level
        """ 
        raise NotImplementedError()

    @property
    def first(self):
        """ First child of current Passage 
        
        :rtype: None or Passage
        :returns: None if current Passage has no children,  first child passage if available
        """
        raise NotImplementedError()

    @property
    def last(self):
        """ Last child of current Passage 
        
        :rtype: None or Passage
        :returns: None if current Passage has no children, last child passage if available
        """
        raise NotImplementedError()

    @property
    def children(self):
        """ Children of the passage

        :rtype: OrderedDict
        :returns: Dictionary of chidren, where key are subreferences
        """
        raise NotImplementedError()


class Text(Resource):
    """ A CTS Text """
    def __init__(self, citation=None, metadata=None, **kwargs):
        super(Text, self).__init__(**kwargs)

        self._cRefPattern = MyCapytain.common.reference.Citation()
        if citation is not None:
            self.citation = citation

        if metadata is not None:
            self.metadata = metadata
        else:
            self.metadata = MyCapytain.common.metadata.Metadata()

    def getValidReff(self, level=1, reference=None):
        """ Given a resource, Text will compute valid reffs 
        
        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param passage: Subreference (optional)
        :type passage: Reference
        :rtype: List.basestring
        :returns: List of levels
        """
        raise NotImplementedError()

    def getPassage(self, reference):
        """ Retrieve a passage and store it in the object
        
        :param reference: Reference of the passage
        :type reference: MyCapytain.common.reference.Reference or List of basestring
        :rtype: Passage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """

        raise NotImplementedError()

    def getLabel(self):
        """ Retrieve metadata about the text
        
        :rtype: dict
        :returns: Dictionary with label informations
        """
        raise NotImplementedError()

    @property
    def reffs(self):
        """ Get all valid reffs for every part of the Text

        :rtype: MyCapytain.resources.texts.tei.Citation
        """
        return [reff for reffs in [self.getValidReff(level=i) for i in range(1, len(self.citation) + 1)] for reff in reffs]

    @property
    def citation(self):
        """ Get the lowest cRefPattern in the hierarchy

        :rtype: MyCapytain.common.reference.Citation
        """
        return self._cRefPattern
    
    @citation.setter
    def citation(self, value):
        """ Set the cRefPattern

        :param value: Citation to be saved
        :type value:  MyCapytain.common.reference.Citation
        """
        if isinstance(value, MyCapytain.common.reference.Citation):
            self._cRefPattern = value


from setuptools import setup, find_packages

import MyCapytain

setup(
  name='MyCapytain',
  version=MyCapytain.__version__,
  description='Abstraction of CTS API for Python',
  url='http://github.com/Capitains/MyCapytain',
  author='Thibault Clerice',
  author_email='leponteineptique@gmail.com',
  license='MIT',
  packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
  install_requires=[
    "requests>=2.8.1",
    "six>=1.10.0",
    "lxml==3.4.4",
    "future==0.15.2"
  ],
  tests_require=[
    "mock==1.3.0",
    "xmlunittest>=0.3.2"
  ],
  extras_require = {
    "DOC" : ["Sphinx==1.3.1"]
  },
  test_suite="tests",
  zip_safe=False
)



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.endpoints.cts5
   :synopsis: CTS5 endpoint implementation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
import MyCapytain.endpoints.proto
import requests


class CTS(MyCapytain.endpoints.proto.CTS):

    """ 
        Basic integration of the MyCapytain.endpoints.proto.CTS abstraction
    """

    def __init__(self, endpoint, inventory=None):
        """ API Prototype object

        :param self: Object
        :type self: API
        :param endpoint: URL of the API
        :type endpoint: str
        :param inventory: Inventory to use
        :type inventory: str

        :ivar endpoint: Url of the endpoint
        :ivar inventory: Default Inventory
        """
        super(CTS, self).__init__(endpoint)
        self.inventory = inventory

    def call(self, parameters):
        """ Call an endpoint given the parameters
        
        :param parameters: Dictionary of parameters
        :type parameters: dict
        :rtype: text
        """
        # DEV !
        parameters = {
            key: str(parameters[key]) for key in parameters if parameters[key] is not None
        }
        if self.inventory is not None and "inv" not in parameters:
            parameters["inv"] = self.inventory

        request = requests.get(self.endpoint, params=parameters)
        return request.text

    def getCapabilities(self, inventory=None):
        """ Retrieve the inventory information of an API 
        
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "request": "GetCapabilities"
        })

    def getValidReff(self, urn, inventory=None, level=None):
        """ Retrieve valid urn-references for a text
        
        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param level: Depth of references expected
        :type level: int
        :return: XML Response from the API as string
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "level": level,
            "request": "GetValidReff"
        })

    def getFirstUrn(self, urn, inventory=None):
        """ Retrieve the first passage urn of a text
        
        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetFirstUrn"
        })

    def getPrevNextUrn(self, urn, inventory=None):
        """ Retrieve the previous and next passage urn of one passage
        
        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetPrevNextUrn"
        })

    def getLabel(self, urn, inventory=None):
        """ Retrieve informations about a CTS Urn
        
        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetLabel"
        })

    def getPassage(self, urn, inventory=None, context=None):
        """ Retrieve a passage
        
        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "context": context,
            "request": "GetPassage"
        })

    def getPassagePlus(self, urn, inventory=None, context=None):
        """ Retrieve a passage and informations about it
        
        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "context": context,
            "request": "GetPassagePlus"
        })



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.endpoints.proto
   :synopsis: Prototypes of APIs endpoint

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""


class API(object):
    """
    API Prototype object

    :param self: Object
    :type self: API
    :param endpoint: URL of the API
    :type endpoint: text

    :ivar endpoint: Url of the endpoint
    """

    def __init__(self, endpoint):
        """ Instantiate an API class
        """
        self.endpoint = endpoint


class Ahab(API):
    """
    Abstract Capitains Ahab API
    See : http://capitains.github.io/pages/ahab.html
    """
    def search(self, query, urn, start=1, limit=5, format="json"):
        """ Perform a search on given namespace

        :param query: Term to perform search on
        :type query: text
        :param urn: Partial or complete urn identifying the request
        :type urn: text
        :param start: Starting element to display
        :type start: int
        :param limit: Limit of result displayed
        :type limit: int
        :param format: Format to request (json or xml)
        :type format: str
        :rtype: str
        """
        raise NotImplementedError()

    def permalink(self, urn, format="xml"):
        """ Perform a permalink request on API

        :rtype: str
        """
        raise NotImplementedError()


class CTS(API):
    """
    CTS API Endpoint Prototype 
    """
    def getCapabilities(self, inventory):
        """ Retrieve the inventory information of an API 

        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getValidReff(self, urn, inventory, level=1):
        """ Retrieve valid urn-references for a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param level: Depth of references expected
        :type level: int
        :rtype: str
        """
        raise NotImplementedError()

    def getFirstUrn(self, urn, inventory):
        """ Retrieve the first passage urn of a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getPrevNextUrn(self, urn, inventory):
        """ Retrieve the previous and next passage urn of one passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getLabel(self, urn, inventory):
        """ Retrieve informations about a CTS Urn

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getPassage(self, urn, inventory, context=None):
        """ Retrieve a passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        raise NotImplementedError()

    def getPassagePlus(self, urn, inventory, context=None):
        """ Retrieve a passage and informations about it
        
        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        raise NotImplementedError()